#! usr/bin/perl

#########################################################
# PROGRAM: TreeRot v3					#
#########################################################							
# Original Author/Architect: 	Dr. Michael Sorenson	#
# Perl Implementation:	 	Eric Franzosa		#
#########################################################

# When the program is run as an executable, this section
# forces the terminal/command window to navigate to the
# program's location for finding other files.  Otherwise
# it may open in a default directory.

use File::Basename;
$dir = dirname($0);
chdir($dir);

#Print the program's heading to the screen

print 	 
	'+---------------------------------------------------------------+',"\n",
	'|                                                               |',"\n",
	'|   TreeRot.v3                                                  |',"\n",
	'|                                                               |',"\n",
	'|   Copyright (c) 2008 Michael D. Sorenson & Eric A. Franzosa   |',"\n",
	'|   Department of Biology, Boston University                    |',"\n",
	'|                                                               |',"\n",
	'+---------------------------------------------------------------+',"\n"
;

print
	 "\nPlease choose: (1) make command file, (2) parse PAUP results: "
;

#prompt for a choice of action
chomp($choice = <STDIN>);
while($choice != 1 and $choice != 2){
	print "Please pick option 1 or 2 to continue: \n";
	chomp($choice = <STDIN>);
}

#call a subroutine to execute the desired function
if($choice == 1){
	&make_command_file;
} else {
	&parse_output_file;
}

#everything is done, prompt user to press a key to exit.
#this prevents the window from closing automatically when
#the program's execution is finished.

print "\nTo exit, press ENTER or just close the window.\n";
exit if <STDIN>;

#########################################################
# This Subroutine Makes a Command File (Option #1)	#
#########################################################

sub make_command_file{

	#########################################################
	# File I/O						#
	#########################################################

	#TRY TO OPEN A TREE FILE TO READ IN
	$tree_file = &open_read_file("\nName of existing tree file?: ");
	open(IN,"$tree_file");
		
	#TRY TO OPEN A COMMAND FILE TO PRINT TO
	$command_file = &open_write_file("\nName for command file?: ");
	open(OUT,">$command_file");	
	
	#ASK FOR A NAME FOR THE PAUP OUTPUT FILE
	#we don't open this here, but the name is included in the command file
	$log_file = &open_write_file("\nName for PAUP log file?: ");
	
	#########################################################
	# Retrieve Possible Partition Information		#
	#########################################################
	
	print "\nIs your dataset partitioned? (Y/N) ";
	chomp($part_answer = <STDIN>);
	
	#boolean $PARTITIONED stores the logical answer to this
	#question (1 for partitioned, 0 otherwise)
	$PARTITIONED = 0;
	
	if($part_answer =~ m/^Y/i){
		#Ask the user for a list of partitions; build this
		#into the array @partition_names
		
		$PARTITIONED = 1;
		
		#how many paritions?
		print "\nNumber of partitions?: ";
		chomp($part_number = <STDIN>);
		
		while($part_number == 0 | $part_number =~ m/[^\d]/){
			print "Not a valid number, try again: ";
			chomp($part_number = <STDIN>);
		}
	
		#Once a valid number is entered, ask for a name for each
		#partition and and store it in @partition_names
		for($i = 1 ; $i <= $part_number ; $i++){
			print "\n\tType a name for partition \#$i: ";
			chomp($this_name = <STDIN>);
			push(@partition_names,$this_name);
		}
		
	}
				
	#########################################################
	# Read in the data and convert formats (if needed)	#
	#########################################################
	
	#this format uses the unusual ^M or \r carriage return 
	#character.  This block will normalize the input for the
	#parsing section.
	
	while($line = <IN>){
		if($line =~ m/\r/){
			@data = split(/\r/,$line);
			last;
		} else {
			chomp($line);
			push @data,$line;
		}
	}
	
	#########################################################
	# Get rid of bracketed comments				#
	#########################################################
	
	#for storing editted lines
	@temp_data = ();
	#boolean: true if we're in a bracket block
	$ENCLOSED = 0;
	
	foreach $line (@data){
		#the only case where we NEED brackets is in the rooted/unrooted declaration
		#protect this case by replacing the brackets with braces {}'s
		$line =~ s/\[(\&.)\]/\{$1\}/;
	
		#remove any complete statement on the line
		#enclosed by brackets
		$line =~ s/\[.*\]//g;
		
		#is there a closing bracket?
		if($line =~ m/\]/){
			#remove everything leading up to the bracket
			$line =~ s/.*\]//;
			#finish this enclosure block
			$ENCLOSED = 0;
		}
		
		#is there an opening bracket?
		elsif($line =~ m/\[/){
			#remove everything after the bracket
			$line =~ s/\[.*//;
			#start an enclosure block
			$ENCLOSED = 1;
		}
		
		#we aren't starting or ending a block, but we are inside one
		elsif($ENCLOSED == 1){
			#kill the line
			$line = '';
		}
		
		else{
			#do nothing, this is a good line
		}
		
		#if there's anything left of the line, keep it
		if($line ne ''){
			push(@temp_data,$line);
		}
	}
	
	#replace the old data lines with the editted ones
	@data = @temp_data;
	
	#########################################################
	# Parse the remaining code for the tree & trans. table  #
	#########################################################
		
	foreach $line (@data){	
		#if you find a line that contains 'tree' and an equals sign (=) and
		#ends with a semicolon (;) ==> this is the tree definition line in PAUP
		if($line =~ m/tree(.*)=(.*)\;/){
			#see if there is a rooted(R)/unrooted(U) definition, if not
			#then default the status to unrooted
			
			#note that &R or &U are now enclosed by {}'s in computer memory
			#after the [] removal process
			
			$line =~ m/\{\&(.)\}/;
			if($1 eq 'R'){
				$type = 'R';
			} else {
				$type = 'U';
			}
			#pull out the actual tree topology
			$line =~ m/(\(.*\))/;
			$tree = $1;
			last;
		}
		#if you find the word 'Translate' then we 
		#need to make a translation table;
		#adjust to case insensitive match for different PAUP versions
		elsif($line =~ m/translate/i){
			$TABLE = 1;
		}
		#if the translation table has begun, add tranlation
		#lines to the table
		elsif($TABLE == 1 and $line =~ m/\s+(\d*) ([^\,\s]*)/){
			$taxon_number = $1;
			$taxon_name   = $2;
			
			$translate{$taxon_number} = $taxon_name;
		}
		else{	
			next;
		}
	}
	
	#########################################################
	# Format Tree: Use Trans. table / remove branch lengths #
	#########################################################
	
	#remove branch lengths from the taxon names, if they are present
	#(\d*) matches whole number component, (\.\d*)? matches possible
	#decimal component
	$tree =~ s/\:(\d*)(\.\d*)?//g;
	
	#translate taxon names, if applicable
	if(defined %translate){
		$tree =~ s/(\d*)/$translate{$1}/g;		
	}

	#########################################################
	# Pull nodes out of the tree	           	 	#
	#########################################################
	
	#Match the smallest node in the tree
	while($tree =~ m/\((.*)\(([^\(\)]*?)\)(.*)\)/){	
		#Break the current tree into three pieces	
		$left  = $1;
		$node  = $2;
		$right = $3;
		
		#remove the ()'s from these blocks
		$left  =~ s/[\(|\)]//g;
		$right =~ s/[\(|\)]//g;
		
		#join the blocks to produce a constraint
		$constraint = "\($left,$right,\($node\)\)";
	 
		#fix formatting problems (double commas, empty Left/Right)
		$constraint =~ s/[,]+/,/g;
		$constraint =~ s/\(,/\(/g;
		$constraint =~ s/,\)/\)/g;
		
		#add this constraint to a list
		push @constraints,$constraint;
		
		#collapse the current node into its parent
		#by removing parentheses
		$tree =~ s/\($node\)/$node/;
	}
	
	close IN;
	
	#########################################################
	# Special options for rooted/unrooted cases      	#
	#########################################################
	
	if($type eq 'U'){
		#if this is an unrooted tree, drop last constraint
		pop @constraints;
	}
			
	#########################################################
	# Prepare the Command File				#
	#########################################################
	
	#OPENING SYNTAX
		print OUT 	"\#NEXUS\n\n";
		print OUT 	"BEGIN PAUP\;\n";
	
	#Print the list of constraints
	for($i = 1 ; $i <= scalar (@constraints) ; $i++){
		print OUT 	"CONSTRAINTS node$i = $constraints[$i-1]\;\n";
	} 
	
	#MIDDLE SYNTAX
		print OUT 	"\n";
		print OUT 	"set autoclose\;\n";
		print OUT	"log file=$log_file\;\n";
		
		#(Partitioned case only)
		if($PARTITIONED == 1){
			print OUT 	"include all\;\n";
		}
		
		print OUT	"hsearch noenforce addseq = random nreps = 20\;\n";
		print OUT	"[!-----*****-----NO CONSTRAINTS-----*****-----]\n";
	
	#Here there is a difference between the partitioned and unpartitioned cases
	#For the partitioned case we must print several lines of PAUP code for each
	#partition.
		
	if($PARTITIONED == 1){
	     	for($i = 0 ; $i <= $#partition_names ; $i++){
	     		     	
	     		print OUT	"exclude all\;\n";
			print OUT	"include $partition_names[$i]\;\n";
			print OUT	"lenfit\;\n";
		}
		
	} else {
			print OUT "lenfit\;\n"	
	}
	
	#Print the individual constraint commands.  Again, we print multiple lines
	#per partition in the partitioned case.

	for($i = 1 ; $i <= scalar (@constraints) ; $i++){

		print OUT 	"\n";

		#(Partitioned case only)
		if($PARTITIONED == 1){
			print OUT 	"include all\;\n";
		}
			
		print OUT	"hsearch enforce converse constraints = node$i addseq = random nreps = 20\;\n";
	
		#change the formatting of the constraint
		$constraints[$i-1] =~ s/\,\(/\,\>\-\-\-\-\-\</;
		$constraints[$i-1] =~ s/[\(|\)]//g;
	
		#print the constraint command
		print OUT	"\[\!$constraints[$i-1]\]\n";
		print OUT	"\n";
		
		#(Partitioned case only)
		if($PARTITIONED == 1){
			for($j = 0 ; $j <= $#partition_names ; $j++){
	     			print OUT	"exclude all\;\n";
				print OUT	"include $partition_names[$j]\;\n";
				print OUT	"lenfit\;\n";
			}
		
		#(Unpartitioned case)
		} else {
				print OUT	"lenfit\;\n";
		}

	}
	
	#CLOSING SYNTAX
		
	#(partitioned case only)(?)
	if($PARTITIONED == 1){
		print OUT 	"\ninclude all\;\n";
	}	
		
	print OUT	"\nlog stop\;\n";
	print OUT	"end\;\n";
	
	close OUT;
	
	print "\n\<COMMAND FILE CREATED\>\n";
} 

#########################################################
# This Subroutine Parses the Output File		#
#########################################################

sub parse_output_file{
	
	#########################################################
	# File I/O						#
	#########################################################
	
	#TRY TO OPEN THE PAUP OUTPUT FILE
	$paup_file = &open_read_file("\nName of existing PAUP output file?: ");
	open(IN,"$paup_file");
		
	#COME UP WITH A NAME FOR THE PARSED OUTPUT FILE

	$parsed_file = &open_write_file("\nName for parsed output file?: ");
	open(OUT,">$parsed_file");

	#########################################################
	# Format Conversion					#
	#########################################################
	
	#this format uses the unusual ^M or \r carriage return 
	#character.  This block will normalize the input for the
	#parsing section.
	
	while($line = <IN>){
		if($line =~ m/\r/){
			@data = split(/\r/,$line);
			last;
		} else {
			chomp($line);
			push @data,$line;
		}
	}

	#########################################################
	# Text Preprocessing					#
	#########################################################
	
	#Preprocess the PAUP textual output.  This gets rid of a
	#lot of the unnecessary material that we don't want to be
	#worrying about in the ACTUAL processing later.	
	
	#process the data lines to make sure that constraints end 
	#on a single line. we assume that constraint statements 
	#contain commas separating the taxa. As long as we see
	#comma lines we glue them together into "constraints."
	#We will check that they are ACTUALLY constraints in the
	#next step. Stop gluing when a blank line appears.
	
	@data_temp = ();
	$constraint = '';
	for($i = 0 ; $i <= $#data ; $i++){
		if($data[$i] =~ m/\,/){
			while($data[$i] ne ''){
				$data[$i] =~ s/\s//g;
				$constraint .= $data[$i];
				$i++;
			}
			push(@data_temp,$constraint);
			$constraint = '';
		}
		push(@data_temp,$data[$i]);
	}	
	
	#clear data (will be refilled next)
	@data = ();
	
	#Make sure that things that LOOKED like constraints were
	#actual constraints.  Get rid of everything else except
	#for tree length data.
	for($line_index = 0; $line_index <= $#data_temp ; $line_index++){
		if($data_temp[$line_index] =~ m/\,/){
			#potential constraint
			if($data_temp[$line_index] =~ m/\>\-\-\-\-\-\</){
				push(@data,$data_temp[$line_index]);
				next;
			} 
		}
		
		#'tree #' signals tree length data
		if($data_temp[$line_index] =~ m/tree \#/i){
			push(@data,$data_temp[$line_index]);
			push(@data,$data_temp[$line_index+1]);
			$line_index++;
		}
	}
	
	#Finally, compound the tree length data in case it falls on 
	#multiple lines (all the trees for one partition should fall
	#on the same line).
	@data_temp = @data;
	@data = ();
	
	$tree_length_line = 'Length ';
	for($line_index = 0; $line_index <= $#data_temp ; $line_index++){
		#if we are starting a constraint, add the old tree length
		#line, start a new one, and add the constraint statement
		if($data_temp[$line_index] =~ m/\>\-\-\-\-\-\</i){
			push(@data,$tree_length_line);
			$tree_length_line = 'Length ';
			push(@data,$data_temp[$line_index]);
		}	
	
		#if we find a new tree length line...
		elsif($data_temp[$line_index] =~ m/^tree \#/i){
			@tree_numbers = split(/\s+/,$data_temp[$line_index]);
			@tree_lengths = split(/\s+/,$data_temp[$line_index+1]);
			
			# if the tree numbering starts at 1 but we already have
			# a longer tree length line, we must be starting a new
			# parition; print the old one
			if($tree_numbers[2] == 1 & $tree_length_line ne 'Length '){
				push(@data,$tree_length_line);
				$tree_length_line = 'Length ';
			}
			
			#remove the word "Length" from the length line we found
			shift @tree_lengths;
			#add the numbers to the tree length line
			$tree_length_line .= "@tree_lengths ";
			
			#jump over the length data that we just looked at
			$line_index++;
		}
	}

	#add the last line of tree lengths
	push(@data,$tree_length_line);
		
	#########################################################
	# Debug Preprocessing					#
	#########################################################
	
	#Uncomment this code to generate a file containing the
	#intermediate output.  This will help diagnose changes
	#in PAUP syntax that are causing parsing problems.
		  	
	#open(TEMP,">temp_preprocess.txt");
	#foreach $line (@data){
	#	print TEMP $line."\n";
	#}
	#close TEMP;

	#########################################################
	# Output Processing					#
	#########################################################
	
	#these pieces store the halves of the full tree; used in the labeling process
	$master_left  = '';
	$Master_right = '';
		
	#store the lines of output until they are ready to print
	@out_lines = ();
			
	#file header
	push(@out_lines,"UNCONSTRAINED CASE(S):\n");
	
	#determine the number of constraints
	$number_of_constraints = scalar(grep(/\>\-\-\-\-\-\</,@data));
	#determine the number of partitions
	$number_of_partitions  = (scalar grep(/Length/,@data))/($number_of_constraints+1);
	
	#store tree length info. for the unconstrained case
	@unconstrained_trees_avg = ();	
	@unconstrained_trees_min = ();
	@unconstrained_trees_max = ();
	
	#build the uncontrainted case
	foreach $p (0..$number_of_partitions-1){
		$line = shift(@data);
		
		#get simple stats from subroutine
		($avg,$max,$min) = &simple_line_stats($line);
		
		#store these stats index is partition number, array format
		push @unconstrained_trees_avg,$avg;
		push @unconstrained_trees_min,$min;
		push @unconstrained_trees_max,$max;
		
		push(@out_lines,sprintf("\tAVG = %.2f\tMIN = %.2f\tMAX = %.2f\n",($avg,$min,$max)));
	}
		
	#deal with all the constraints
	foreach $c (1..$number_of_constraints){
		#first, process the constraint statement
		$line = shift(@data);
		$line =~ m/(.*)\,\>\-\-\-\-\-\<(.*)/;
		
		#grab the node name
		$node 	  = "$2";
		
		#keep track of the "full tree" for labeling purposes later
		#basing it on the biggest constaint ensures that everything collapses properly
		if(length($node)>length($master_right)){
			$master_left  = $1;
			$master_right = $2;
		}
		
		#store right and left taxa for removal of redundant nodes (later)
		push(@left, $1);
		push(@right,$2);
		
		#print new node heading
		push(@out_lines,"\n$node\n");
		
		#now go through the partitions
		foreach $p (1..$number_of_partitions){
			$line = shift(@data);
		
			#get simple stats from subroutine
			($avg,$max,$min) = &simple_line_stats($line);
		
			#store these stats (index is partition number, array format
			push(@out_lines,
			sprintf("\tAVG = %.2f (%.2f)\tMIN = %.2f (%.2f)\tMAX = %.2f (%.2f)\n",
				(	
				$avg, $avg-$unconstrained_trees_avg[$p-1],
				$min, $min-$unconstrained_trees_max[$p-1],
				$max, $max-$unconstrained_trees_min[$p-1]
				)
			));
			
			#store data for labeled tree in a hash
			#{key1 = partition, one for each tree}{key2 = node} = decay index 
			$nodes{$p-1}{$node} = sprintf("%.2f",$avg-$unconstrained_trees_avg[$p-1]);
		}
	}			
	
	#########################################################
	# Identify Redundancy Among the Nodes we processed	#
	#########################################################
	
	#All constraints are of the form (A,(B)).  If a pair
	#of contraints (A,(B)) and (B,(A)) exists, drop one of
	#them to eliminate redundancy.
	
	#if there is a left half of the constraint that looks like this
	#right half, mark that the right half (a node) is redundant. Note
	#that the loop indexing prevents us from killing BOTH participants
	#in a redundant pair.
		
	#identify redundancy, work from the bottom up		
	for($i = $#left ; $i >= 0 ; $i--){
		for($j = 0 ; $j <= $i ; $j++){
			if($left[$i] eq $right[$j]){
				$redundant{$right[$j]} = 1;
			}
		}
	}
			
	#########################################################
	# Make the labeled Tree					#
	#########################################################
	
	print OUT "\#NEXUS\n\n";
	print OUT "Begin trees\;\n\n";
	
	#print a tree for each partition
	
	foreach $i (sort {$a <=> $b} keys %nodes){
		#add the full tree to the set of nodes
		$nodes{$i}{"$master_left,$master_right"} = 1;

		#as long as there is something OTHER than the full tree left in the
		#nodes hash, there is something left to collapse...
		while(scalar(keys %{$nodes{$i}}) > 1){
		
			#########################################################################
			#Sort the remaining nodes.  The full tree always winds up at the top  	#
			#[0] because it is longest.  The next largest node always winds up at 	#
			#position [1].  We must collapse in decreasing order of size in order 	#	   
			#to use s/// to find overlaps.						#	
			#									#
			# Example 								#
			#   ABC,BC,C : ABC =~ s/BC/(BC) ==> A(BC) C  | A(BC) =~ s/C/(C)/    	#
			#	==> A(B(C)) check!						#	
			# Counterexample							#	
			#   ABC,C,BC : ABC =~ s/C/(C)/  ==> AB(C) BC | AB(C) =~ s/BC/(BC)/  	#
			#	==> does not work!						#
			#########################################################################

			@sorted_nodes = sort {length($b) <=> length($a)} keys %{$nodes{$i}};
			
			#find the longest node (the full tree)
			$full_tree = $sorted_nodes[0];
			#the next node we will try to collapse is the next biggest node
			$next_node = $sorted_nodes[1];		
	
			#remember the name of the full tree (we'll use this as a pointer to 
			#the original value after we've changed the name of the longest node)
			$temp_node = $full_tree;
			
			#collapse the smaller node into the full tree (adding surrounding syntax)
			#(1) if this is a redundant node, don't label it:
			if($redundant{$next_node} == 1){		
				$full_tree =~ s/$next_node/\($next_node\)/;
			
			} else {
			#(2) do label it
				$full_tree =~ s/$next_node/\($next_node\)\[\&decayindex\=$nodes{$i}{$next_node}\]/;
			}
			
			#assign the old value of the full tree to the new full tree
			$nodes{$i}{$full_tree} = $nodes{$i}{$temp_node};
			#delete the collapsed node and the old full tree from the hash
			delete $nodes{$i}{$temp_node};
			delete $nodes{$i}{$next_node};
		}
		
		#when we exit the loop there is only one node left in %nodes, the full tree
		@sorted_nodes = sort {length($b) <=> length($a)} keys %{$nodes{$i}};
		$tree = $sorted_nodes[0];	
		
		#print out the formatted tree to the file
		$tree_number = $i+1;
		print OUT "tree partition$tree_number \= \($tree\)\;\n";
	}
	
	print OUT "\nEnd\;\n\n";
	
	#########################################################
	# Print out all the other information to the file	#
	#########################################################
	
	#stuff inside of [ ]'s will be ignored
	print OUT "\[\n";
	
	#don't print the information for a redundant node
	$REDUNDANT = 0;
	
	foreach $line (@out_lines){
		#to compare nodes in %redundant we need to eliminate the \n char
		$potential_node = $line;
		$potential_node =~ s/\n//g;
		
		if($potential_node =~ m/\,/){
			#found a comma, indicates that we are considering a new
			#constraint statement.  Forget what you were "thinking"
			#about redundancy
			$REDUNDANT = 0;
		}
		
		if(defined($redundant{$potential_node})){
			#this is the start of a redundant block of information
			$REDUNDANT = 1;
		}
		
		if($REDUNDANT != 1){		
			print OUT $line;
		}
	}
	
	print OUT "\]\n";
	
	#DONE!
	
	print "\n\<PARSING OF RESULTS COMPLETE\>\n";
	
	close IN;
	close OUT;		
} 

#########################################################
# Simple Statistics Used in the Parsing Subroutine	#
#########################################################

sub simple_line_stats{
	#find the average length of the trees
  	#also keep track of the min and max length
  		
  	my($line_to_process) = @_;
  	@values = split(/\s+/,$line);
  		
  	#get rid of the word "length"
  	shift @values;

	#begin statistics
	$avg = 0;
	$max = $values[0];
	$min = $values[0];

  	foreach $e (@values){
  		#add component to average
  		$avg += $e/(scalar @values);
  		#check min
  		if($e < $min){
  			$min = $e;
  		}
  		#check max
  		if($e > $max){
  			$max = $e;
  		}
  	}
  	
  	return ($avg,$max,$min);
}

#########################################################
# Reusable File Opening Subroutine #1			#
#########################################################

sub open_write_file{
	($instructions) = @_; 	
	print "$instructions";
	$SUCCESS = 0;
	while($SUCCESS == 0){
		chomp($file_name = <STDIN>);
		#does this file already exists?
		if(open(TEST,"$file_name")){
			print "A file with that name already exists.  Overwrite? (Y/N) ";
			chomp($overwrite_answer = <STDIN>);
			#yes, overwrite?
			if($overwrite_answer =~ m/^Y/i){
				close TEST;
				$SUCCESS = 1;
			} 
			#no, try again
			else {
				print "\nPlease choose another file name: ";
			}
		} 
		#file does not already exist
		else {
			$SUCCESS = 1;
		}
	}
	return $file_name;	
}

#########################################################
# Reusable File Opening Subroutine #2			#
#########################################################

sub open_read_file{
	($instructions) = @_; 	
	print "$instructions";
	$SUCCESS = 0;
	while($SUCCESS == 0){
		chomp($file_name = <STDIN>);
		if(open(TEST,"$file_name")){
			#file opened successfully
			close TEST;
			$SUCCESS = 1;
		} else {
			print "\nFile not found, please try again: ";
		}
	}
	return $file_name;
}
 